"""Fixtures for test modules"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import io
from json import loads
from os import makedirs
from os.path import join, expanduser
from shutil import rmtree
from unittest.mock import patch

from future import standard_library
from gobble.snapshot import SNAPSHOTS_DIR
from pytest import fixture

from gobble import FiscalDataPackage
from gobble.config import settings, GOBBLE_MODE
from gobble.config import ROOT_DIR

standard_library.install_aliases()


# Devlopment mode required
# -----------------------------------------------------------------------------
class BadTestingConfiguration(Exception):
    pass


if GOBBLE_MODE != 'Development':
    raise BadTestingConfiguration()


# Fiscal data packages
# -----------------------------------------------------------------------------

@fixture(scope='session')
def fiscal_package():
    file = join(ROOT_DIR, 'assets', 'datapackage', 'datapackage.json')
    return FiscalDataPackage(file)


@fixture(scope='session')
def invalid_fiscal_package():
    file = join(ROOT_DIR, 'assets', 'datapackage', 'invalid.json')
    return FiscalDataPackage(file)


# Fake the user's local set-up
# -----------------------------------------------------------------------------

@fixture
def tmp_user_dir(request):
    original = settings.USER_DIR
    settings.USER_DIR = join(expanduser('~'), '.gobble.tmp')
    try:
        makedirs(settings.USER_DIR)
    except IOError:
        pass

    def switch_back():
        settings.USER_DIR = original
        try:
            rmtree(settings.USER_DIR)
        except IOError:
            pass

    request.addfinalizer(switch_back)


# User authentication and permission
# -----------------------------------------------------------------------------
DUMMY_DIR = join(ROOT_DIR, 'assets', 'dummy')


@fixture
def token():
    filepath = join(DUMMY_DIR, 'token.json')
    with io.open(filepath) as file:
        return loads(file.read())


@fixture
def permissions():
    filepath = join('', 'permissions.json')
    with io.open(filepath) as file:
        return loads(file.read())


# Mock Gobble and site-package classes
# -----------------------------------------------------------------------------
@fixture
def mock_package():
    package = patch('datapackage.DataPackage', autospec=True)
    package.descriptor = {'name': 'mexican-budget-samples'}
    snapshot_file = join(SNAPSHOTS_DIR, 'POST.datastore.json')
    with io.open(snapshot_file) as file:
        package.description = loads(file.read())['request_json']
    return package


# noinspection PyShadowingNames
@fixture
def mock_user():
    return patch('gobble.user.User', autospec=True)


@fixture
def mock_batch():
    return patch('gobble.uploader.FiscalDataPackage', autospec=True)


@fixture
def mock_uploader():
    return patch('gobble.uploader.Uploder', autospec=True)



"""Tests for the user maodule"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import responses
from future import standard_library
from requests import Response

from gobble.api import authenticate_user

standard_library.install_aliases()


@responses.activate
def test_create_user_hits_correct_api_endpoint():
    responses.add(
        authenticate_user.method,
        authenticate_user.url,
    )
    assert isinstance(authenticate_user(), Response)
    assert authenticate_user().status_code == 200



"""This module exposes user configurable settings"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from logging import DEBUG
from os import getenv
from os import mkdir
from os.path import isdir
from os.path import join, abspath, dirname, expanduser
from sys import modules

from future import standard_library

standard_library.install_aliases()

HOME_DIR = abspath(join(expanduser('~')))
ROOT_DIR = abspath(join(dirname(__file__), '..'))
GOBBLE_MODE = getenv('GOBBLE_MODE', 'Production')


class Production(object):
    OS_URL = 'https://next.openspending.org'
    S3_BUCKET_URL = 'https://s3.amazonaws.com/datastore.openspending.org'
    LOCALHOST = ('127.0.0.1', 8001)
    USER_DIR = join(HOME_DIR, '.gobble')
    LOG_FILE = join(HOME_DIR, '.gobble', 'gobble.log')
    EXPANDED_LOG_STYLE = True
    CONSOLE_LOG_LEVEL = None
    FILE_LOG_LEVEL = DEBUG
    CONSOLE_LOG_FORMAT = '[%(name)s] [%(levelname)s] %(message)s'
    FILE_LOG_FORMAT = ('[%(name)s] '
                       '[%(asctime)s] '
                       '[%(module)s] '
                       '[%(funcName)s] '
                       '[%(levelname)s] '
                       '%(message)s')


class Development(Production):
    S3_BUCKET_URL = 'http://fakes3/fake-bucket'
    OS_URL = 'http://dev.openspending.org'
    USER_DIR = join(HOME_DIR, '.gobble.dev')
    FILE_LOG_LEVEL = DEBUG
    EXPANDED_LOG_STYLE = True
    CONSOLE_LOG_LEVEL = None
    CONSOLE_LOG_FORMAT = ('[%(name)s] '
                          '[%(module)s] '
                          '[%(funcName)s] '
                          '[%(levelname)s] '
                          '%(message)s')


settings = getattr(modules[__name__], GOBBLE_MODE)

if not isdir(settings.USER_DIR):
    mkdir(settings.USER_DIR)



"""Tests for the uploader module"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


from datapackage.exceptions import ValidationError
from future import standard_library
from os.path import join
from pytest import raises

# noinspection PyUnresolvedReferences
from tests.fixtures import fiscal_package, invalid_fiscal_package
from gobble.config import ROOT_DIR
from gobble.fiscal import compute_hash


standard_library.install_aliases()


DATA_FILE = join(ROOT_DIR, 'assets', 'datapackage', 'data', 'data.csv')


def test_compute_hash_returns_correct_hash():
    assert compute_hash(DATA_FILE) == '+uqBmwvQLi0M2W2enNxD/A=='


# noinspection PyShadowingNames
def test_validate_bad_package_with_raise_error_false(invalid_fiscal_package):
    report = invalid_fiscal_package.validate(
        raise_on_error=False,
        schema_only=True
    )
    assert isinstance(report, list)
    assert 'is a required property' in report[0]


# noinspection PyShadowingNames
def test_validate_bad_package_raises_error_by_default(invalid_fiscal_package):
    with raises(ValidationError):
        invalid_fiscal_package.validate()


# noinspection PyShadowingNames
def test_validate_returns_true_for_valid_package(fiscal_package):
    report = fiscal_package.validate(
        raise_on_error=False,
        schema_only=True
    )
    assert isinstance(report, list)


# def test_report_validation_errors_logs_error_messages(capsys):
#     bad_package = DataPackage({'foo': 'bar'})
#     report_validation_errors(bad_package)
#     stdout, stderr = capsys.readouterr()
#     print('validating')
#     print(settings.CONSOLE_LOG_LEVEL)
#     print('yo', stdout)
#     if settings.CONSOLE_LOG_LEVEL <= INFO:
#         assert 'validating' in stdout
#         assert 'required property' in stdout



""" This modules uploads data-packages to the Open-Spending datastore"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import io
import sys

from base64 import b64encode
from hashlib import md5
from io import StringIO
from os.path import getsize, join, basename, isfile
from time import sleep
from json import dumps
from datapackage import DataPackage
from datapackage.exceptions import ValidationError
from future import standard_library
from gobble.user import User
from goodtables.pipeline import Pipeline
from requests import HTTPError
from requests_futures.sessions import FuturesSession

from gobble.config import settings
from gobble.logger import log
from gobble.api import (handle,
                        upload_package,
                        request_upload,
                        toggle_publish,
                        upload_status)

standard_library.install_aliases()


HASHING_BLOCK_SIZE = 65536
OS_DATA_FORMATS = ['.csv']
POLL_PERIOD = 5
REPORT_FILENAME = 'goodtables.report.json'


class ToggleError(Exception):
    pass


class UploadError(Exception):
    pass


def compute_hash(filepath):
    """Return the md5 hash of a file"""
    hasher = md5()

    with io.open(filepath, 'rb') as stream:
        chunk = stream.read(HASHING_BLOCK_SIZE)
        while len(chunk) > 0:
            hasher.update(chunk)
            chunk = stream.read(HASHING_BLOCK_SIZE)

    md5_binary = hasher.digest()
    md5_bytes = b64encode(md5_binary)
    md5_unicode = md5_bytes.decode('utf-8')

    return md5_unicode


class FiscalDataPackage(DataPackage):
    """This class represents a fiscal data package.

    The class is a subclass of the :class:`datapackage.DataPackage` class.
    The constructor takes the same arguments as its parent class, except that
    the schema is "fiscal".

    :param target: The target is the full path to the fiscal datapackage JSON
    descriptor, but it can also be a dictionary representing the schema itself
    or a url pointing to a descriptor (for more information please refer to the
    documentation for the :class:`datapackage.DataPackage` class.
    :param user: a `gobble.user.user` object.
    """

    def __init__(self, filepath, user=None, **kw):
        if not isfile(filepath):
            raise NotImplemented('%s is not a local path', filepath)

        super(FiscalDataPackage, self).__init__(filepath,
                                                schema='fiscal', **kw)
        self._check_file_formats()

        self._streams = []
        self._session = FuturesSession()
        self._futures = []
        self._responses = []

        self.user = user
        self.name = self.descriptor.get('name')
        self.path = basename(filepath)
        self.filepath = filepath

    def validate(self, raise_on_error=True, schema_only=True):
        """Validate a datapackage schema.

        By default, only the data-package schema is validated. To validate the
        data files too, set the `data` flag to `True`. The method fails if an
        error is found, unless the `raise_error` flag is explicitely set to
        `False`.

        :param raise_on_error: raise error on failure or not (default: True)
        :param schema_only: only validate the schema (default: True)
        :raise: :class:`ValidationError` if the schema is invalid

        :return A list of error messages or an empty list.
        """
        messages = []

        if raise_on_error:
            super(FiscalDataPackage, self).validate()
        else:
            try:
                super(FiscalDataPackage, self).validate()
                message = '%s (%s) is a valid fiscal data-package schema'
                log.info(message, self.path, self)

            except ValidationError:
                for error in self.iter_errors():
                    message = 'SCHEMA ERROR in %s: %s'
                    args = self.path, error.message
                    messages.append(message % args)
                    log.warn(message, *args)

        if messages:
            messages.append('Aborting data validation due to invalid schema')
            return messages

        if not schema_only:
            return self._validate_data(raise_on_error)
        else:
            return messages

    def upload(self, publish=True, skip_validation=False):
        """Upload a fiscal datapackage to Open-Spending.

        It does this in 3 steps:
            * request upload urls for AWS S3 storage
            * upload all files to the owner's S3 bucket
            * insert the data into the Open-Spending datastore (PostgreSQL)

        By default, newly uploaded packages are kept private, but you can
        change that with the `publish` flag. Also note that if you upload the
        same fiscal data package again, the previous version will be
        overwritten.

        For now, the only valid datafile format is CSV.

        :param skip_validation: use only if you have already done so
        :param publish: toggle the datapackage to "published" after upload
        """
        self.descriptor['author'] = self.user.name
        self.descriptor['owner'] = self.user.id

        with io.open(self.filepath, 'w') as descriptor:
            descriptor.write(self.to_json())

        if not skip_validation:
            self.validate()

        log.info('Starting uploading process for %s', self)

        for s3_target in self._request_s3_upload():
            self._push_to_s3(*s3_target)

        self._handle_promises()
        self._insert_into_datastore()

        sleep(POLL_PERIOD)
        while self.in_progress:
            sleep(POLL_PERIOD)

        if publish:
            self.toggle('public')

        return self.url

    @property
    def url(self):
        return join(settings.OS_URL, self.user.id + ':' + self.name)

    @property
    def in_progress(self):
        """Return true when the upload finished."""

        query = dict(datapackage=self._descriptor_s3_url)
        answer = upload_status(params=query).json()
        args = self, answer['status'], answer['progress'], len(self)
        log.debug('%s is loading (%s) %s/%s', *args)
        if answer['status'] == 'fail':
            raise UploadError(answer.get('error'))
        
        return answer['status'] not in {'done', 'fail'}

    def toggle(self, to_state):
        """Toggle public access to a fiscal datapackage

        Change the status of a fiscal data package from public to private or
        vice-versa. If something went wrong, whilst changing the status, you
        will get a :class:`upload.ToggleError`.

        :param to_state: the unique name of the datapackage
        :return: the new state of the package, i.e. "public" or "private"
        """
        publish = True if to_state == 'public' else False
        package_id = self.user.id + ':' + self.name
        query = dict(
            jwt=self.user.permissions['os.datastore']['token'],
            id=package_id,
            publish=publish
        )
        answer = handle(toggle_publish(params=query))

        if not answer['success']:
            message = 'Unable to toggle datapackage to %s'
            raise ToggleError(message, to_state)

        log.info('%s is now %s', package_id, to_state)
        return to_state

    def _check_file_formats(self):
        for resource in self:
            if resource.descriptor['mediatype'] != 'text/csv':
                message = 'Usupported format: %s, valid formats are %s'
                raise NotImplemented(message, resource.path, OS_DATA_FORMATS)

    def _validate_data(self, raise_on_error):
        """Validate the package resources with GoodTables."""

        def summarize(feedback_, path_):
            intro = 'GoodTables has detected some errors in %s.' % path_
            hint = 'Please check out the full report: %s.' % REPORT_FILENAME

            info = feedback_['meta']
            summary = (
                'There are {bad_rows} (out of {total_rows}) bad rows '
                'and {bad_cols} (out of {total_cols}) bad columns. '
            ).format(
                bad_rows=info['bad_row_count'],
                total_rows=info['row_count'],
                bad_cols=info['bad_column_count'],
                total_cols=len(info['columns'])
            )

            log.debug(intro + summary)
            return [intro, summary, hint]

        for resource in self:
            schema = resource.descriptor['schema']
            path = resource.descriptor['path']
            filepath = join(self._base_path, path)

            pipeline = Pipeline(filepath, report_stream=StringIO())
            pipeline.register_processor('schema', options={'schema': schema})
            is_valid, report = pipeline.run()

            if is_valid:
                return []

            if raise_on_error:
                raise ValidationError('%s is invalid' % filepath)
            else:
                feedback = report.generate()
                with open(REPORT_FILENAME, 'w+') as json:
                    json.write(dumps(feedback, indent=4, ensure_ascii=False))
                return summarize(feedback, path)

    @property
    def filedata(self):
        filedata = {
            resource.descriptor['path']: {
                'name': resource.descriptor['name'],
                'length': getsize(resource.local_data_path),
                'md5': compute_hash(resource.local_data_path),
                'type': resource.descriptor['mediatype'],
            } for resource in self
        }
        descriptor_file = {
            basename(self.filepath): {
                'name': self.name,
                'length': getsize(self.filepath),
                'md5': compute_hash(self.filepath),
                'type': 'application/octet-stream',
            }
        }
        filedata.update(descriptor_file)
        return {
            'filedata': filedata,
            'metadata': {
                'owner': self.user.id,
                'name': self.name
            }
        }

    @property
    def bytes(self):
        return sum([file['length']
                    for file in self.filedata['filedata'].values()])

    def _get_header(self, path, content_type):
        filepath = join(self.base_path, path)
        return {'Content-Length': str(getsize(filepath)),
                'Content-MD5': compute_hash(filepath),
                'Content-Type': content_type}

    @property
    def _descriptor_s3_url(self):
        return join(settings.S3_BUCKET_URL, self.user.id, self.name, self.path)

    def _request_s3_upload(self):
        """Request AWS S3 upload urls for all files.
        """
        response = request_upload(
            params=dict(jwt=self.user.permissions['os.datastore']['token']),
            json=self.filedata
        )
        files = handle(response)['filedata']

        for path, info in files.items():
            message = '%s is ready for upload to %s'
            log.info(message, path, info['upload_url'])
            query = {k: v[0] for k, v in info['upload_query'].items()}
            yield (info['upload_url'],
                   path,
                   query,
                   self._get_header(path, info['type']))

    def _push_to_s3(self, url, path, query, headers):
        """Send data files for upload to the S3 bucket.
        """

        log.debug('Started uploading %s to %s', path, url)
        log.debug('Headers: %s', headers)
        log.debug('Query parameters: %s', query)

        absolute_path = join(self.base_path, path)
        stream = io.open(absolute_path, mode='rb')
        future = self._session.put(url,
                                   headers=headers,
                                   data=stream,
                                   params=query,
                                   background_callback=self._s3_callback)

        self._streams.append(stream)
        self._futures.append(future)

    @staticmethod
    def _s3_callback(_, response):
        handle(response)
        log.info('Successful S3 upload: %s', response.url)

    def _handle_promises(self):
        """Collect all promises from S3 uploads.
        """
        for stream, future in zip(self._streams, self._futures):
            exception = future.exception()
            if exception:
                raise exception
            response = future.result()

            if response.status_code != 200:
                message = 'Something went wrong uploading %s to S3: %s'
                log.error(message, response.url, response.text)
                raise HTTPError(message)

            self._responses.append(response)
            stream.close()

    def _insert_into_datastore(self):
        """Transfer datafiles from S3 into the postgres datastore.

        :return: the url of the fiscal datapackage on Open-Spending
        """
        query = {
            'jwt': self.user.permissions['os.datastore']['token'],
            'datapackage': self._descriptor_s3_url
        }
        response = upload_package(params=query)
        handle(response)

        log.info('Congratuations, %s was uploaded successfully!', self)
        log.info('You can find you fiscal datapackage here: %s', self.url)

        return self.url

    def __len__(self):
        return len(self.resources)

    def __repr__(self):
        return '<FiscalDataPackage [%s files]: %s>' % (len(self), self.name)

    def __str__(self):
        return self.name

    def __iter__(self):
        for resource in self.resources:
            yield resource

    def __getitem__(self, index):
        return self.resources[index]


if __name__ == '__main__':
    user_ = User()
    filepath_ = sys.argv[1]
    package_ = FiscalDataPackage(filepath_, user=user_)
    package_.upload(publish=True)



"""Test the API module."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import responses
from future import standard_library
from pytest import mark, raises
from requests import HTTPError

from gobble.api import handle
from tests.parameters import api_calls

standard_library.install_aliases()


@responses.activate
@mark.parametrize('api_call', api_calls)
def test_api_calls_hit_correct_endpoints(api_call):
    responses.add(api_call.method, api_call.url)
    assert api_call().status_code == 200


@responses.activate
@mark.parametrize('api_call', api_calls)
def test_handle_returns_json_if_status_is_200(api_call):
    responses.add(api_call.method, api_call.url, body='{"foo": "bar"}')
    assert handle(api_call()) == {'foo': 'bar'}


@responses.activate
@mark.parametrize('api_call', api_calls)
def test_handle_raises_error_if_status_is_400(api_call):
    responses.add(api_call.method, api_call.url, body=HTTPError())
    with raises(HTTPError):
        handle(api_call())


# @responses.activate
# @mark.parametrize('api_call', api_calls)
# def test_call_endpoint_with_query_parameters(api_call):
#     responses.add(
#         api_call.method,
#         api_call.url + '?spam=eggs&foo=bar',
#         match_querystring=False
#     )
#     params = dict(foo='bar', spam='eggs')
#     assert api_call(params=params).status_code == 200



"""A simple logger that logs to file and console"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from future import standard_library
from logging import (getLogger,
                     FileHandler,
                     StreamHandler,
                     DEBUG,
                     Formatter,
                     Filter)

from gobble.config import settings, GOBBLE_MODE

standard_library.install_aliases()


class MultilineFilter(Filter):
    """Split a multi-line message over several log records"""

    # http://stackoverflow.com/questions/22934616
    # This is not considered good practice but it
    # preserves indentation and improves readability.
    # I log the same information twice anyway.

    # TODO: display the correct (calling) module in the log record

    def filter(self, record):
        message = record.getMessage()

        if '\n' not in message:
            return True

        for line in message.split('\n'):
            log.debug(line)
        return False


def _configure_logger(name):
    logger = getLogger(name)
    multiline = MultilineFilter()

    logger.addFilter(multiline)
    logger.setLevel(DEBUG)

    if settings.FILE_LOG_LEVEL:
        file = FileHandler(settings.LOG_FILE)
        file.setLevel(settings.FILE_LOG_LEVEL)
        file.setFormatter(Formatter(settings.FILE_LOG_FORMAT))
        logger.addHandler(file)

    if settings.CONSOLE_LOG_LEVEL:
        stream = StreamHandler()
        stream.setLevel(settings.CONSOLE_LOG_LEVEL)
        stream.setFormatter(Formatter(settings.CONSOLE_LOG_FORMAT))
        logger.addHandler(stream)

    return logger


log = _configure_logger('Gobble')


log.debug('Gobble is running in %s mode', GOBBLE_MODE)


for key, value in vars(settings).items():
    if key.isupper():
        log.debug('%s = %s', key, value)



"""Test the logger module"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from logging import FileHandler, StreamHandler, Logger, LogRecord
from future import standard_library

standard_library.install_aliases()

from gobble.logger import log, MultilineFilter


def test_gobble_logger_exists_and_has_correct_name():
    assert isinstance(log, Logger)
    assert log.name == 'Gobble'


def test_logger_has_multiline_filter():
    any(list(map(lambda x: isinstance(x, MultilineFilter), log.filters)))


def test_logger_has_at_least_one_handler():
    file = any(list(map(lambda x: isinstance(x, FileHandler), log.handlers)))
    cons = any(list(map(lambda x: isinstance(x, StreamHandler), log.handlers)))
    assert any((file, cons))


def test_multifile_filter_catches_multiline_records():
    multiline = LogRecord('foo', 0, 'bar', 0, '\n', (), 'spam')
    MultilineFilter().filter(multiline) is False



"""This module handles all API calls"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import str
from requests import HTTPError
from requests import Session
from urllib.parse import (urlencode,
                          urljoin,
                          urlunsplit,
                          urlsplit)
try:
    from json import JSONDecodeError
except ImportError:
    try:
        from simplejson.scanner import JSONDecodeError
    except ImportError:
        JSONDecodeError = ValueError

from gobble.config import settings
from gobble.logger import log
from gobble.snapshot import SnapShot, to_json


# One session for all queries
_session = Session()


class EndPoint(object):
    """An API endpoint served by the os-conductor app"""

    def __init__(self, method, *path,
                 trailing_slash=False,
                 session=_session):

        self.has_slash = trailing_slash
        self.method = method
        self._path = path
        self._session = session
        self.snapshot = None

    @property
    def fire(self):
        return getattr(self._session, self.method.lower())

    @property
    def url(self):
        return urljoin(settings.OS_URL, '/'.join(self.path))

    @property
    def path(self):
        """Deal with trailing slashes
        """
        path = list(self._path)
        if self.has_slash:
            path[-1] += '/'
        return path

    def _append(self, params):
        """Encode the query string and append it to the endpoint
        """
        query = urlencode(tuple(params.items()))
        parts = urlsplit(self.url)._replace(query=query)
        return urlunsplit(parts)

    def __call__(self, params=None, **request):
        """Fire, take a snapshot and return the response
        """
        request_url = self._append(params) if params else self.url
        response = self.fire(request_url, **request)
        summary = request_url, response, params
        self.snapshot = SnapShot(self, *summary, **request)
        return response

    def __str__(self):
        return '%s: /%s' % (self.method, '/'.join(self.path))

    def __repr__(self):
        return '<EndPoint ' + str(self) + '>'

    @property
    def info(self):
        return {
            'method': self.method,
            'path': self.path,
            'endslash': self.has_slash,
            "url": self.url
        }


# Expose callable endpoint objects to other modules
# -----------------------------------------------------------------------------
authenticate_user = EndPoint('GET', 'user', 'check')
authorize_user = EndPoint('GET', 'user', 'authorize')
oauth_callback = EndPoint('GET', 'oauth', 'callback')
update_user = EndPoint('POST', 'user', 'update')
search_packages = EndPoint('GET', 'search', 'package')
upload_package = EndPoint('POST', 'package', 'upload')
upload_status = EndPoint('GET', 'package', 'status')
request_upload = EndPoint('POST', 'datastore', trailing_slash=True)
toggle_publish = EndPoint('POST', 'package', 'publish')
# -----------------------------------------------------------------------------


def handle(response):
    """Handle a response and return its payload

    If all is okay, return the json payload of the response
    as a dict. If the status code is in the range 400 to 599,
    raise an HTTPError.
    """
    try:
        response.raise_for_status()
    except HTTPError as error:
        log.error(error)
        raise error

    try:
      return to_json(response)
    except JSONDecodeError:
        return {}



"""Test the snapshot module."""

from pytest import mark


# noinspection PyShadowingNames
from gobble.snapshot import freeze
from tests.parameters import json_objects


@mark.parametrize('unsecure_json, secure_json', json_objects)
def test_remove_secrets_does_its_job(unsecure_json, secure_json):
    freeze(unsecure_json)
    assert unsecure_json == secure_json


# @responses.activate
# @mark.parametrize('function', api_functions)
# def test_endpoint_saves_request_snapshot_to_file(function):
#     function()
#     assert isfile(function.snapshot._filepath)
#     assert function.snapshot._filepath.endswith('json')
# # noinspection PyShadowingNames,PyUnusedLocal



"""Manage user authentication and authorization"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import io
import os

from builtins import dict
from future import standard_library
from future.backports.http.server import HTTPServer, SimpleHTTPRequestHandler
from os.path import join
from json import loads
from os import mkdir
from threading import Thread
from json import dumps

from gobble.logger import log
from gobble.config import settings
from gobble.api import (handle,
                        authenticate_user,
                        authorize_user,
                        oauth_callback)

standard_library.install_aliases()


OS_SERVICES = ['os.datastore']
TOKEN_FILE = join(settings.USER_DIR, 'token.json')
PERMISSIONS_FILE = join(settings.USER_DIR, 'permission.json')
AUTHENTICATION_FILE = join(settings.USER_DIR, 'authentication.json')
TOKEN_ENV_VAR = 'GOBBLE_AUTH_TOKEN'


class InvalidToken(Exception):
    pass


class User(object):
    """A contributor on Open-Spending.

    Note: you don't need to instantiate or even import this class yourself.
    The correlation is that Gobble only supports a single user for now.

    If you use Gobble for the first time, run the :method:`user.start` function
    first. It will help you obtain your authentication token. Open-Spending
    uses Google OAuth2, so you will need a valid Google email to make it work.

    Logs, tokens, permissions and snapshots of the latest API requests are
    saved in the user directory, which by default is :data:`~/.gobble`.
    """

    def __init__(self):
        self.folder = settings.USER_DIR
        self.token = self._uncache_token()
        self.authentication = self._request_authentication()
        self.permissions = self._request_permissions()
        self.id = self.authentication['profile']['idhash']
        self.name = self.authentication['profile']['name']

    @classmethod
    def uncache(cls, key):
        filepath = join(settings.USER_DIR, key + '.json')
        with io.open(filepath) as cache:
            json = loads(cache.read())
            log.debug('Your %s is %s', key, json)
            return json

    @staticmethod
    def _uncache_token():
        """Read the token from the cache.
        """
        if TOKEN_ENV_VAR in os.environ:
            return os.environ[TOKEN_ENV_VAR]
        with io.open(TOKEN_FILE) as cache:
            json = loads(cache.read())
            log.debug('Your token is %s', json['jwt'])

            return json['jwt']

    def _request_authentication(self):
        """Ask Open-Spending if the token is valid.
        """
        query = dict(jwt=self.token)
        response = authenticate_user(params=query)
        authentication = handle(response)

        if not authentication['authenticated']:
            message = 'Token has expired: request a new one'
            log.error(message)
            raise InvalidToken(message)

        name = authentication['profile']['name']
        log.info('Hello %s! You are logged into Open-Spending', name)
        return authentication

    def _request_permissions(self):
        """Request permissions for Open-Spending services.
        """
        permissions = {}

        for service in OS_SERVICES:
            query = dict(jwt=self.token, service=service)
            response = authorize_user(params=query)
            permission = handle(response)
            permissions.update({service: permission})

            return permissions

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<User: ' + str(self) + '>'


class LocalHost(SimpleHTTPRequestHandler):
    """A local server to catch and save the token."""

    def do_GET(self):
        log.debug('Callback received: %s', self.path)
        token = self.path[6:]

        if token:
            with io.open(TOKEN_FILE, 'w+', encoding='utf-8') as file:
                file.write(dumps({'token': token}, ensure_ascii=False))
                log.info('Saved your token in %s', TOKEN_FILE)


def create_user():
    """Obtain the new user a token.

    Open-Spending uses Google OAuth2 for authentication, so you will need a
    valid Google email address to make this work. When you click on the link
    provided, you will be redirected to your browser to sign up. That's it.
    """

    def install_user_folder():
        try:
            mkdir(settings.USER_DIR)
        except FileExistsError:
            pass

    def request_new_token():
        localhost = 'http://%s:%s' % settings.LOCALHOST
        query = dict(next=localhost, callback_url=oauth_callback.url)
        response = authenticate_user(query)
        authorization = handle(response)
        new_thread = Thread(target=listen_for_token)
        prompt_user(authorization)
        local_server = new_thread.run()
        local_server.join()
        log.info('Well done, you have a now token!')

    def prompt_user(authorization):
        sign_up_url = authorization['providers']['google']['url']
        message = ('Please open a new private browsing window '
                   'and paste this link to get a token: %s')
        log.critical(message, sign_up_url)

    def listen_for_token():
        server = HTTPServer(settings.LOCALHOST, LocalHost)
        server.serve_forever()

    def cache(info_, file):
        with io.open(file, 'w+', encoding='utf-8') as json:
            json.write(dumps(info_, ensure_ascii=False, indent=4))

    install_user_folder()
    request_new_token()

    user_ = User()

    cached_info = (
        (user_.token, TOKEN_FILE),
        (user_.permissions, PERMISSIONS_FILE),
        (user_.authentication, AUTHENTICATION_FILE)
    )
    for info in cached_info:
        cache(*info)
        log.debug('Cached %s to %s', *info)

    return user_


# Expose a user object for use in other modules
# user = User()



"""Test the configuration module"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from urllib.parse import urlparse

from future import standard_library
from os.path import isdir, dirname
from pytest import mark

standard_library.install_aliases()

from gobble.config import ROOT_DIR, HOME_DIR, GOBBLE_MODE
from tests.parameters import configurations


@mark.parametrize('settings', configurations)
def test_log_levels_are_set(settings):
    assert settings.CONSOLE_LOG_LEVEL is None or \
           isinstance(settings.CONSOLE_LOG_LEVEL, int)


@mark.parametrize('settings', configurations)
def test_log_levels_are_set(settings):
    assert isinstance(settings.LOG_FILE, str)


def test_global_constants_are_correct():
    assert GOBBLE_MODE in ('Production', 'Development', 'Testing')
    assert isdir(HOME_DIR)
    assert isdir(ROOT_DIR)


@mark.parametrize('settings', configurations)
def test_user_and_log_folders_exist(settings):
    assert isdir(settings.USER_DIR)
    if settings.LOG_FILE:
        assert isdir(dirname(settings.LOG_FILE)) or None


@mark.parametrize('settings', configurations)
def test_os_url_is_parsable(settings):
    assert urlparse(settings.OS_URL).scheme
    assert urlparse(settings.OS_URL).netloc


@mark.parametrize('settings', configurations)
def test_log_formats_are_strings(settings):
    assert isinstance(settings.CONSOLE_LOG_FORMAT, str)
    assert isinstance(settings.FILE_LOG_FORMAT, str)


@mark.parametrize('settings', configurations)
def test_localhost_is_a_netloc_tuple(settings):
    assert isinstance(settings.LOCALHOST[0], str)
    assert isinstance(settings.LOCALHOST[1], int)


@mark.parametrize('settings', configurations)
def test_expanded_log_style_setting_is_set(settings):
    assert isinstance(settings.EXPANDED_LOG_STYLE, bool)



"""Search for users and package inside the ElasticSearch database"""


from gobble.api import search_packages, handle
from gobble.user import User


ALLOWED_KEYS = {
    'title': 'title',
    'author': 'author',
    'description': 'description',
    'region': 'regionCode',
    'country': 'countryCode',
    'city': 'cityCode'
}


def search(global_query=None, keyed_query=None, private=True, limit=None):
    """Query the packages on Open-Spending.

    You can search a package by `title`, `author`, `description`, `regionCode`,
    `countryCode` or`cityCode`. You can match all these fields at once with the
    magic `q` key.

    If authentication-token was provided, then private packages from the
    authenticated user will also be included. Otherwise, only public packages
    will be returned. You can limit the size of your results with the `size`
    parameter.

    :param global_query: an expression to search for across all keys
    :param keyed_query: a `dict` of key/value search pairs
    :param private: search private datapackages or not (default: True)
    :param limit: the number of results returned (default: None)

    :return: a dictionary with the results
    :rtype: :class: `list` of `dict`
    """

    assert global_query or keyed_query

    if keyed_query:
        _check_keys(keyed_query)
        raw_query = _prefix_keys(keyed_query)
    else:
        raw_query = {}

    if global_query:
        raw_query.update(q=global_query)

    query = _quote_values(raw_query)

    if limit:
        query.update(size=limit)

    if private:
        query.update(jwt=User().token)

    response = search_packages(params=query)
    return handle(response)


def _quote_values(query):
    values = ['"' + str(v) + '"' for v in query.values()]
    return dict(zip(query.keys(), values))


def _prefix_keys(query):
    keys = ['package.' + str(ALLOWED_KEYS[k]) for k in query.keys()]
    return dict(zip(keys, query.values()))


def _check_keys(query):
    for key in dict(query).keys():
        if key not in ALLOWED_KEYS:
            msg = 'Invalid search key "{key}"'
            raise ValueError(msg.format(key=key))


if __name__ == '__main__':
    search(global_query='loic', private=True)



"""This module has good intentions, like helping you debug API calls"""


from collections import OrderedDict
from copy import deepcopy
from datetime import datetime
from json import loads, dumps
from os import listdir
from os.path import join, isdir
from re import search
from click import command

import io

from json import JSONDecodeError

from gobble.config import ROOT_DIR
from gobble.logger import log
from gobble.config import settings


SNAPSHOTS_DIR = join(ROOT_DIR, 'assets', 'snapshots')


def to_json(response):
    """Safely extract the payload from the response object"""
    try:
        text = response.text
        return loads(text)
    except JSONDecodeError:
        return {}


class SnapShot(OrderedDict):
    """A chatty wrapper around the API transaction"""

    def __init__(self, endpoint, url, reponse, params,
                 headers=None, json=None, is_freeze=False):
        """Log, record and save before returning an instance"""

        self.is_freeze = is_freeze

        self.url = url
        self.endpoint = endpoint
        self.response = reponse
        self.headers = headers
        self.params = params
        self.request_payload = json

        super(SnapShot, self).__init__(self._template)
        self.timestamp = str(datetime.now())

        self._log()
        self._record()
        self._save()

    def _log(self):
        """Is there such a thing as too much logging?"""

        code = self.response.status_code
        reason = self.response.reason
        response_json = to_json(self.response)
        begin = code, reason, self.endpoint, 'begin'
        end = code, reason, self.endpoint, 'end'
        transaction = ' [%s] %s - %s (%s) '

        log.debug('{:*^100}'.format(transaction % begin))

        messages = (
            ('Request endpoint: %s', self.endpoint.url),
            ('Request time: %s', self.response.elapsed),
            ('Request parameters: %s', self.params),
            ('Request payload: %s', self.request_payload),
            ('Request headers: %s', self.headers),
            ('Response headers: %s', self.response.headers),
            ('Response payload: %s', response_json),
            ('Response cookies: %s', self.response.cookies),
            ('Request full URL: %s', self.url),
        )
        for message in messages:
            log.debug(*message)

        indent = 4 if settings.EXPANDED_LOG_STYLE else None
        log.debug(dumps(response_json, ensure_ascii=False, indent=indent))

        log.debug('{:*^100}'.format(transaction % end))

    def _record(self):
        """Store the transaction info"""

        json = to_json(self.response)
        duplicate_json = deepcopy(json)

        self['timestamp'] = self.timestamp
        self['url'] = self.url
        self['query'] = self.params
        self['request_json'] = self.request_payload
        self['response_json'] = duplicate_json
        self['request_headers'] = self.headers
        self['response_headers'] = dict(self.response.headers)
        self['cookies'] = dict(self.response.cookies)

    @property
    def _template(self):
        return (
            ('timestamp', None),
            ('host', settings.OS_URL),
            ('url', None),
            ('method', self.endpoint.method),
            ('path', self.endpoint.path),
            ('query', None),
            ('request_json', None),
            ('response_json', None),
            ('request_headers', None),
            ('response_headers', None),
            ('cookies', None),
        )

    def _save(self):
        """Save the snapshot as JSON in the appropriate place"""
        with io.open(self._filepath, 'w+', encoding='utf-8') as file:
            file.write(dumps(self, ensure_ascii=False))
        log.debug('Saved request + response to %s', self._filepath)

    @property
    def _folder(self):
        return SNAPSHOTS_DIR if self.is_freeze else settings.USER_DIR

    @property
    def _filepath(self):
        template = '{method}.{path}.json'
        dot_path = '.'.join(self.endpoint._path).rstrip('/')
        params = {'method': self.endpoint.method, 'path': dot_path}
        filename = template.format(**params)
        return join(self._folder, filename)

    def __str__(self):
        return str(self.endpoint) + ' at ' + self.timestamp

    def __repr__(self):
        return '<SnapShot %s>' % str(self)

    @property
    def json(self):
        return dumps(self, ensure_ascii=False)


def freeze(json):
    """Recursively substitute unwanted strings inside a json-like object

    Basically, remove anything in the substitution list below, even when
    hidden in inside query strings.
    """
    subs = {
        'jwt': r'jwt=([^&^"]+)',
        "bucket_id": r'\/([\w]{32})\/',
        'Signature': r'Signature=([^&^"]+)',
        'AWSAccessKeyId': r'AWSAccessKeyId=([^&^"]+)',
        'Expires': r'Expires=([^&^"]+)',
        'Date': None,
        "Set-Cookie": None,
        'token': None,
    }

    def regex(dummy_, json_, key_, pattern_, value_):
        match = search(pattern_, value_)
        if match:
            sub = match.group(1), dummy_
            json_[key_] = value_.replace(*sub)

    if isinstance(json, list):
        for item in json:
            freeze(item)
    elif isinstance(json, dict):
        for field, pattern in subs.items():
            for key, value in json.items():
                dummy = field.upper()
                if key == field:
                    json[key] = dummy
                elif isinstance(value, str):
                    if pattern:
                        regex(dummy, json, key, pattern, value)
                elif isinstance(value, dict):
                    freeze(value)


@command
def archive(destination):
    """Freeze and move all snapshots to the destination folder."""

    if not isdir(destination):
        raise NotADirectoryError(destination)

    for file in listdir(settings.USER_DIR):
        verb = file.split('.')[0]
        if verb in ['GET', 'POST', 'PUT']:

            with io.open(file) as source:
                snapshot = loads(source.read())

            freeze(snapshot)

            # Overwrite if necessary
            output = join(destination, file)
            with io.open(output, 'w+', encoding='utf-8') as target:
                target.write(dumps(snapshot, ensure_ascii=False))



"""The command line interface for Gobble"""

import io
import sys
from collections import defaultdict
from json import loads, dumps
from os.path import join, splitext, basename
from shutil import rmtree

from datapackage.exceptions import ValidationError
from gobble.config import ROOT_DIR, settings
from gobble.fiscal import FiscalDataPackage
from gobble.search import search as elasticsearch, ALLOWED_KEYS
from gobble.user import create_user, User
from click import (Choice,
                   command,
                   argument,
                   Path,
                   group,
                   option,
                   echo,
                   secho,
                   edit,
                   version_option)


DEFAULT_STYLE = dict(fg='white', bold=True)
ERROR_STYLE = dict(fg='red', bold=True)
SUCCESS_STYLE = dict(fg='green', bold=True)
ITEM_STYLE = dict(fg='yellow')

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


# Main command
# -----------------------------------------------------------------------------

def _get_package(key):
    filepath = join(ROOT_DIR, 'package.json')
    with io.open(filepath) as json:
        package = loads(json.read())
        return package[key]


@group(context_settings=CONTEXT_SETTINGS)
@version_option(
    version=_get_package('version'),
    prog_name=_get_package('slug'),
    is_flag=True,
)
def gobble():
    pass


# View
# -----------------------------------------------------------------------------

_FILES = [
    'log',
    'package',
    'settings',
    'token',
    'authentication',
    'permissions'
]


@command(
    epilog='FILE: %s' % ', '.join(_FILES),
    context_settings=CONTEXT_SETTINGS
)
@argument(
    'file',
    type=Choice(_FILES),
)
@option(
    '-e', '--editor',
    help='Open the file inside an editor.',
    is_flag=True
)
def view(file, editor):
    """Display (or edit) various Gobble files.

    Use the --editor option if you want to open the file inside an editor.
    Note that only the settings file is meant to be manually edited.
    """
    extension = '.log' if file == 'log' else '.json'
    file = 'gobble' if file == 'log' else file
    folder = ROOT_DIR if file == 'package' else settings.USER_DIR
    filepath = join(folder, file + extension)

    if editor:
        saved = edit(filename=filepath)
        if saved:
            secho('Saved changes to %s' % filepath, **SUCCESS_STYLE)

    else:
        try:
            with io.open(filepath) as cache:
                echo(cache.read())
        except FileNotFoundError:
            secho('%s could not be found' % filepath, **ERROR_STYLE)


# User
# -----------------------------------------------------------------------------

_ACTIONS = ['create', 'delete']


@command(context_settings=CONTEXT_SETTINGS)
@argument(
    'action',
    required=True,
    type=Choice(_ACTIONS),
    metavar='[' + '|'.join(_ACTIONS) + ']'
)
def user(action):
    """Create or delete a user.

    To create a new Gobble user, you need to be registered online. Gobble only
    supports a single user at a time. A work-around to switch user accounts is
    to delete and recreate a new user. Just make sure that you copy-paste the
    authorization link inside a private browser window when prompted.
    """
    if action == 'create':
        echo('Welcome to Open-Spending %s' % create_user())
    elif action == 'delete':
        rmtree(settings.USER_DIR)

        try:
            username = User.uncache('authentication')['username']
        except FileNotFoundError:
            username = 'user unknown'

        args = (settings.USER_DIR, username)
        secho('Deleted %s (%s)' % args, **DEFAULT_STYLE)


# Validate
# -----------------------------------------------------------------------------

@command(context_settings=CONTEXT_SETTINGS)
@argument(
    'filepath',
    required=True,
    type=Path(exists=True, resolve_path=True)
)
@option(
    '-s', '--schema',
    help='Validate only the schema.',
    is_flag=True
)
def validate(filepath, schema):
    """Validate a fiscal data-package.

    The FILEPATH is the relative or absolute path to a datapackage. By default,
    the command validates both the schema and the data files.
    """
    _, extension = splitext(filepath)
    filename = basename(filepath)

    if extension != '.json':
        secho('Expecting %s to be JSON file' % filepath, **ERROR_STYLE)
        sys.exit(1)

    package = FiscalDataPackage(filepath)
    error_messages = package.validate(raise_on_error=False, schema_only=schema)

    if error_messages:
        for error_message in error_messages:
            secho(error_message, **ERROR_STYLE)
    else:
        secho('SUCCESS! %s is valid' % filename, **SUCCESS_STYLE)


# Upload
# -----------------------------------------------------------------------------

@command(context_settings=CONTEXT_SETTINGS)
@argument(
    'filepath',
    type=Path(exists=True, resolve_path=True),
    required=True
)
@option(
    '-s', '--skip',
    help='Skip validation (not recommended).',
    is_flag=True
)
@option(
    '-p', '--private',
    help='Keep the online data-package private.',
    is_flag=True
)
def upload(filepath, private, skip):
    """Upload a fiscal package to Open-Spending.

    The FILEPATH is the relative or absolute path to the data-package file.
    The data is always validated before upload, unless validation is skipped
    explicitely. This is not recommended. Once uploaded, the data-package will
    published, unless you pass the --private flag.
    """

    user_ = User()
    package = FiscalDataPackage(filepath, user=user_)

    args = package, len(package), package.bytes
    message = 'Uploading %s to Open-Spending (%s files, %s bytes)...'
    secho(message % args, **DEFAULT_STYLE)

    try:
        url = package.upload(publish=not private, skip_validation=skip)

        state = 'privately' if private else 'publicly'
        args = package, state, url
        message = 'Congratulations! %s is now %s available online: %s.'

        secho(message % args, **SUCCESS_STYLE)

    except ValidationError:
        message = (
            'Upload aborted due to invalid data schema or data file. \n'
            'Use the validate command to find out more about the problem.'
        )
        secho(message, **ERROR_STYLE)


# Search
# -----------------------------------------------------------------------------

@command(context_settings=CONTEXT_SETTINGS)
@argument(
    'expression',
    required=False
)
@option(
    '-w', '--where',
    help='Search a specific field (may be repeated).',
    type=(Choice(list(ALLOWED_KEYS.keys())), str),
    nargs=2,
    metavar='KEY VALUE',
    multiple=True
)
@option(
    '-l', '--limit',
    help='Limit the number of results.',
    default=False,
    type=int
)
@option(
    '-p', '--private',
    help='Include private packages belonging to the user.',
    is_flag=True
)
@option(
    '-r', '--raw',
    help='Return the results as JSON.',
    is_flag=True
)
def search(expression, where, limit, private, raw):
    """Search fiscal packages on Open-Spending.

    You can search for an EXPRESSION across all keys or restrict yourself to
    specific keys using the --where option. You can also combine both
    approaches. Allowed search keys are:

    \b
        - title,
        - author
        - description
        - region
        - country
        - city

    The command returns a short summary of each matching fiscal package. To
    return the complete search results as JSON, pass the --raw flag. Results
    do not include the private packages of the current user by default.
    """
    results = elasticsearch(
        global_query=expression,
        keyed_query=dict(where),
        limit=limit,
        private=private
    )
    template = ('Country: {countryCode} \n'
                'Begin: {begin} \n'
                'End: {end} \n'
                'Author: {author} \n'
                'ID: {id}')

    if not raw:
        if results:
            for result in results:
                package = result['package']
                package.update(id=result['id'])

                if 'fiscalPeriod' in package:
                    begin = package['fiscalPeriod'].get('start', 'unknown')
                    end = package['fiscalPeriod'].get('end', 'unknown')
                    package.update(begin=begin)
                    package.update(end=end)

                info = defaultdict(lambda: 'unknown', **result['package'])
                url = '/'.join([settings.OS_URL, 'viewer', result['id']])

                secho('Title: %s' % package['title'], **ITEM_STYLE)
                secho(template.format(**info), **DEFAULT_STYLE)
                secho('View: %s' % url)
                secho('Download: %s \n' % result['origin_url'])

            summary = '%s packages matched your criteria.'
            secho(summary % len(results), **SUCCESS_STYLE)

        else:
            message = 'No online packages matched your criteria.'
            secho(message, **ERROR_STYLE)

    else:
        secho(dumps(results, ensure_ascii=False, indent=4))


gobble.add_command(validate)
gobble.add_command(upload)
gobble.add_command(user)
gobble.add_command(view)
gobble.add_command(search)



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from json import loads
from os.path import join, dirname
from setuptools import setup, find_packages


def read(filename):
    path = join(dirname(__file__), filename)
    with open(path, 'rt') as file:
        return file.read()


def clean(file):
    for line in file.splitlines():
        if line and not line.startswith('-r'):
            yield line


# Prepare
readme = read('README.md')
license_ = read('LICENSE.txt')
requirements = list(clean(read('requirements.txt')))
requirements_dev = list(clean(read('requirements.dev.txt')))
package = loads(read('package.json'))

# Run
setup(
    name=package['name'],
    version=package['version'],
    description=package['description'],
    long_description=readme,
    author=package['author'],
    author_email=package['author_email'],
    url=package['repository'],
    license=package['license'],
    include_package_data=True,
    packages=find_packages(exclude=['specs', 'tests']),
    package_dir={package['slug']: package['slug']},
    install_requires=requirements,
    tests_require=requirements_dev + requirements,
    test_suite='nose.collector',
    zip_safe=False,
    keywords=package['keywords'],
    classifiers=package['classifiers'],
    entry_points={'console_scripts': ['gobble = gobble.cli:gobble']}
)



"""Test the ElasticSearch module"""


import responses
from pytest import raises
from requests import Response

from gobble.search import _check_keys, _quote_values, _prefix_keys
from gobble.api import search_packages


def test_validate_bad_query_raises_exception():
    with raises(ValueError):
        bad_query = dict(foo='bar', name='mickey mouse')
        _check_keys(bad_query)


def test_quote_values_adds_double_quotes():
    original = {'author': 'mickey', 'title': 'fantasia'}
    validated = {'author': '"mickey"', 'title': '"fantasia"'}
    assert _quote_values(original) == validated


def test_prefix_keys_adds_prefixes():
    original = {'author': 'mickey', 'title': 'fantasia'}
    validated = {'package.author': 'mickey', 'package.title': 'fantasia'}
    assert _prefix_keys(original) == validated


def test_numbers_get_quoted_too():
    original = {'int': 1, 'float': 99.99}
    validated = {'int': '"1"', 'float': '"99.99"'}
    assert _quote_values(original) == validated


@responses.activate
def test_calling_search_function_hits_correct_api_endpoint():
    responses.add(
        search_packages.method,
        search_packages.url,
        body='{"it": "worked"}',
    )
    assert isinstance(search_packages(), Response)
    assert search_packages().status_code == 200
    assert search_packages().json() == {'it': 'worked'}
